using System;
using System.DirectoryServices;
using System.Collections;

namespace InfiniTec.ActiveDirectory
{
	/// <summary>
	/// Helper class for Active Directory access in conjunction with Exchange Server 2000/2003
	/// </summary>
	public class ADLayer
	{
		#region Private member fields

		private string _Server;
		private string _Username;
		private string _Password;
		private AuthenticationTypes _AuthenticationType;
		private DirectoryEntry _ConfigurationNamingContext;
		private DirectoryEntry _RootDse;

		#endregion 

		/// <summary>
		/// Gets or sets the domain controller or global catalog to query.
		/// </summary>
		/// <remarks>
		/// <p>Note that this value can also be filled with a valid domain name. 
		/// This requires that the call is made from a computer in the Active Directory forest.</p>
		/// If set to null, serverless binding will be used.
		/// </remarks>
		public string Server 
		{
			get { return _Server; }
			set { 
				_Server = value; 
				_ConfigurationNamingContext = null;
				_RootDse = null;
			}
		}
		/// <summary>
		/// Gets or sets the username to be used for the access. If set to null, integrated authentication
		/// will be used.
		/// </summary>
		public string Username 
		{
			get { return _Username; }
			set { _Username = value; }
		}

		/// <summary>
		/// Gets or sets the passwrd to be used for the access.
		/// </summary>
		public string Password 
		{
			get { return _Password; }
			set { _Password = value; }
		}

		/// <summary>
		/// Gets or sets the authentication type to be used for the access. See 
		/// <see cref="System.DirectoryServices.AuthenticationTypes"/> for more information.
		/// </summary>
		public string AuthenticationType 
		{
			get { return _AuthenticationType.ToString(); }
			set { _AuthenticationType = (AuthenticationTypes) Enum.Parse(typeof(AuthenticationTypes), value, true); }
		}

		/// <summary>
		/// Initializes a new instance of this class.
		/// </summary>
		/// <param name="server">The <see cref="Server"/> to use</param>
		/// <param name="username">The <see cref="Username"/> to use. If set to null, integrated authentication will be used.</param>
		/// <param name="password">The <see cref="Password"/> to use</param>
		/// <param name="authenticationType">The <see cref="AuthenticationType"/> type to use.</param>
		public ADLayer(string server, string username, string password, string authenticationType): this(server, username, password, false)
		{
			AuthenticationType = authenticationType;
		}

		/// <summary>
		/// Creates a new instance of this class, with integrated authentication enabled, and 
		/// using serverless binding. The connection is initialized as read-only.
		/// </summary>
		public ADLayer(): this(true) {}

		/// <summary>
		/// Creates a new instance of this class, with integrated authentication enabled, and 
		/// using serverless binding. The connection is initialized as read-only.
		/// </summary>
		/// <param name="readOnly">Specifies if the connection should be initialized as read-only</param>
		public ADLayer(bool readOnly): this(null, null, null, readOnly)
		{
		}

		/// <summary>
		/// Initializes a new instance of this class
		/// </summary>
		/// <param name="server">The <see cref="Server"/> to use</param>
		/// <param name="username">The <see cref="Username"/> to use. If set to null, integrated authentication will be used.</param>
		/// <param name="password">The <see cref="Password"/> to use</param>
 		/// <param name="readOnly">Specifies if the connection should be initialized as read-only</param>
		public ADLayer(string server, string username, string password, bool readOnly) 
		{
			Username = username;
			Password = password;
			Server = server;

			_AuthenticationType = AuthenticationTypes.Secure;
			if (readOnly) _AuthenticationType |= AuthenticationTypes.ReadonlyServer;
			if (server != null) _AuthenticationType |= AuthenticationTypes.ServerBind;
		}

		/// <summary>
		/// Initializes a new instance of this class that connects read-only to the specified server,
		/// using integrated authentication.
		/// </summary>
		/// <param name="server"></param>
		public ADLayer(string server): this(server, null, null, true)
		{
		}

		/// <summary>
		/// Retrieves an URL that can be used to issue WebDAV commands against the
		/// mailbox of the given user.
		/// </summary>
		/// <param name="name">Either the Windows logonname or a distinguished name.</param>
		/// <param name="useGC">True, if a Global Catalog should be used for resolving the username.</param>
		/// <returns>The uri of the user</returns>
		public string GetMailboxUri(string name, bool useGC) 
		{
			DirectoryEntry userEntry;
			DirectoryEntry homeMDB;
			DirectoryEntry httpVirtualServerRoot;
			DirectorySearcher searcher;
			SearchResultCollection httpVirtualServer;
			DirectoryEntry defaultSmtpVirtualServer;
			string defaultSmtpDomain;
			string proxyAddress;
			string exServerDn;
			string alias;
			string domain;
			string uri;
			string[] serverBinding;

			if (name == null) throw new ArgumentException("name must not be null", name);
			
			if (name.ToUpper().StartsWith("CN=")) 
			{
				userEntry = GetDirectoryEntry(name);
			}
			else 
			{
				userEntry = FindUser(name, useGC);
			}

			if (userEntry == null) throw new Exception("User object could not be found within the forrest.");

			homeMDB = GetDirectoryEntry((string) userEntry.Properties["homeMDB"].Value);
			exServerDn = (string) homeMDB.Properties["msExchOwningServer"].Value;
			httpVirtualServerRoot = GetDirectoryEntry(string.Concat("CN=http, CN=Protocols, ", exServerDn));

			searcher = new DirectorySearcher(httpVirtualServerRoot, "(objectClass=protocolCfgHttpServer)", new string[] {"msExchServerBindings", "msExchDefaultDomain"}, SearchScope.OneLevel);

			httpVirtualServer = searcher.FindAll();

			defaultSmtpVirtualServer = GetDefaultHttpVirtualServer(httpVirtualServer);

			if (defaultSmtpVirtualServer == null) throw new Exception("Unable to find default http virtual server!");

			defaultSmtpDomain = DefaultSMTPDomain.ToLower();
			uri = "";
			// Iterate through the proxy-address of the user to find the SMTP addresses
			foreach (string address in userEntry.Properties["proxyAddresses"]) 
			{
				proxyAddress = address.ToLower();
				
				// Check wether the found address is a SMTP address and if the 
				// domain-part equals the defaultSmtpDomain.

				if ((proxyAddress.StartsWith("smtp:")) && (proxyAddress.IndexOf(defaultSmtpDomain) != -1)) 
				{
					alias = proxyAddress.Substring(5, proxyAddress.IndexOf("@")-5);
					serverBinding = ((string) defaultSmtpVirtualServer.Properties["msExchServerBindings"][0]).Split(':');

					if (serverBinding[2] != "") uri = string.Format("http://{0}:{1}/exchange/{2}", serverBinding[2], serverBinding[1], alias);
					else if (serverBinding[0] != "") uri = string.Format("http://{0}:{1}/exchange/{2}", serverBinding[0], serverBinding[1], alias);
					else uri = string.Format("http://{0}:{1}/exchange/{2}", GetExchangeServerTcpIpAddress(GetDirectoryEntry(exServerDn)), serverBinding[1], alias);
				}
			}
			if (uri == "") 
			{
				//None of the smtp-addresses matched the defaultSmtp Domain so check the
				//other http virtual server if one is responsible for the domain of the mail address.

				foreach (string address in userEntry.Properties["proxyAddresses"]) 
				{
					proxyAddress = address.ToLower();
					if (!proxyAddress.StartsWith("smtp:")) 
					{
						continue;
					}

					proxyAddress = proxyAddress.ToLower();
					alias = proxyAddress.Substring(5, proxyAddress.IndexOf("@")-5);
					domain = proxyAddress.Substring(proxyAddress.IndexOf("@")+1);

					foreach (SearchResult httpServer in httpVirtualServer) 
					{
						// If the virtual http server directory entry does not contain
						// the property msExchDefaultDomain, it is the default smtp host.
						// That node has been handled before, so ignore it here
						if (!httpServer.Properties.Contains("msExchDefaultDomain")) 
						{
							continue;
						}
						// Check if the value of the property msExchDefaultDomain equals
						// the desired domain.
						if (((string) httpServer.Properties["msExchDefaultDomain"][0]) != domain) 
						{
							continue;
						}

						serverBinding = ((string) httpServer.Properties["msExchServerBindings"][0]).Split(':');

						if (serverBinding[2] != "") uri = string.Format("http://{0}:{1}/{2}", serverBinding[2], serverBinding[1], alias);
						else if (serverBinding[0] != "") uri = string.Format("http://{0}:{1}/{2}", serverBinding[0], serverBinding[1], alias);
						else uri = string.Format("http://{0}:{1}/{2}", GetExchangeServerTcpIpAddress(GetDirectoryEntry(exServerDn)), serverBinding[1], alias);
					}
				}
			}
			return uri;
		}

		/// <summary>
		/// Resolves a NetBIOS name (i.e. CONTOSO) to its distinguished name (i.e. DC=Contoso, DC=local)
		/// </summary>
		/// <param name="netbiosName">The NetBios name to resolve</param>
		/// <returns>The distinguished name of the Active Directory partition that corresponds to the 
		/// specified NetBIOS name</returns>
		/// <exception cref="ArgumentException">Thrown, when the netbios cannot be resolved.</exception>
		public string ResolveNetBiosNameToDN(string netbiosName) 
		{
			try 
			{
				return (string) GetDirectoryEntry(string.Format("CN={0}, CN=Partitions, {1}", netbiosName, (string) ConfigurationNamingContext.Properties["distinguishedName"].Value)).Properties["nCName"].Value;
			} 
			catch (System.Runtime.InteropServices.COMException ex) 
			{
				if ((uint) ex.ErrorCode == 0x80072030) throw new ArgumentException("The given NetBIOS name was invalid", "netbiosName", ex);
				else throw;
			}
		}

		/// <summary>
		/// Searches the specified httpServer collection for the default http server.
		/// </summary>
		/// <param name="httpServer">A list with </param>
		/// <returns>The <see cref="System.DirectoryServices.DirectoryEntry"/> with the default httpServer.</returns>
		private DirectoryEntry GetDefaultHttpVirtualServer(SearchResultCollection httpServer) 
		{
			int i;
			i = 0;

			while ((i<httpServer.Count) && (httpServer[i].Properties.Contains("msExchDefaultDomain"))) i++;

			return (i<httpServer.Count) ? httpServer[i].GetDirectoryEntry() : null;
		}
		/// <summary>
		/// Searches the forrest for the directory entry of the given user. 
		/// </summary>
		/// <param name="principalname">Name of the user to find. Must be in the form domain\user</param>
		/// <param name="useGC">Specifies whether to use the Global Catalog to find the user. If false, a standard LDAP-Query is used.</param>
		/// <returns>The directoryentry of the user, if it is found. Null otherwise.</returns>
		public DirectoryEntry FindUser(string principalname, bool useGC) 
		{
			DirectoryEntry searchRoot;
			DirectorySearcher searcher;
			string[] name;
			string ncName;

			name = principalname.Split('\\');
			if (name.Length != 2) throw new ArgumentException("principalname is not in the correct format", principalname);
			
			ncName = ResolveNetBiosNameToDN(name[0]);
			
			searchRoot = GetDirectoryEntry(ncName, useGC);
			
			searcher = new DirectorySearcher(searchRoot, 
				string.Format("(&(objectCategory=person)(objectClass=user)(sAMAccountName={0}))", name[1]), 
				new string[] {"distinguishedName"}, SearchScope.Subtree);

			try 
			{
				return searcher.FindOne().GetDirectoryEntry();
			} 
			catch (NullReferenceException ex) 
			{
				throw new ArgumentException("The given username was not found", "principalname", ex);
			}
		}

		/// <summary>
		/// Returns the TCP/IP address for the specified server object
		/// </summary>
		/// <param name="serverObject"></param>
		/// <returns></returns>
		public string GetExchangeServerTcpIpAddress(DirectoryEntry serverObject) 
		{
			IList networkAddress;
			int i;

			networkAddress = serverObject.Properties["networkAddress"];
			i = 0;

			while (!((string) networkAddress[i]).StartsWith("ncacn_ip_tcp:")) i++;

			return ((string) networkAddress[i]).Substring(13);
		}

		/// <summary>
		/// Returns an <see cref="ADSearch"/> search object that contains the "All Users" address list.
		/// </summary>
		public ADSearch AllUser 
		{
			get 
			{
				DirectoryEntry root;

				root = GetDirectoryEntry(RootDse.Properties["defaultNamingContext"].Value.ToString());

				return new ADSearch("All User", new DirectorySearcher(root, "(objectClass=user)"));
			}
		}
		/// <summary>
		/// Returns an <see cref="ADSearch"/> search object that contains the "All Groups" address list.
		/// </summary>
		public ADSearch AllGroups
		{
			get 
			{
				DirectoryEntry root;

				root = GetDirectoryEntry(RootDse.Properties["defaultNamingContext"].Value.ToString());

				return new ADSearch("All Groups", new DirectorySearcher(root, "(objectClass=group)"));
			}
		}

		/// <summary>
		/// Returns the DirectoryEntry with the specified distinguished name, using a standard LDAP connection.
		/// </summary>
		/// <remarks>This overload does not open the directory entry with a global catalog</remarks>
		/// <param name="distinguishedName">The URI of the directory entry to open.</param>
		/// <returns>The <see cref="System.DirectoryServices.DirectoryEntry" /> of the specified path.</returns>
		public DirectoryEntry GetDirectoryEntry(string distinguishedName) 
		{
			return GetDirectoryEntry(distinguishedName, false);
		}

		/// <summary>
		/// Returns the DirectoryEntry with the specified distinguished name, using a standard LDAP connection.
		/// </summary>
		/// <param name="distinguishedName">The URI of the directory entry to open.</param>
		/// <param name="useGC">True, if a Global Catalog should be queried. False, otherwise</param>
		/// <returns>The <see cref="System.DirectoryServices.DirectoryEntry" /> of the specified path.</returns>
		public DirectoryEntry GetDirectoryEntry(string distinguishedName, bool useGC) 
		{
			string path;

			path = (useGC) ? "GC://" : "LDAP://";
			if (_Server != null) path += _Server;
			
			if ((!path.EndsWith("/")) && (distinguishedName != "")) path += "/";
			path += distinguishedName;

			if (path.EndsWith("//")) path = path.Remove(path.Length-2, 2);
			return new DirectoryEntry(path, Username, Password, _AuthenticationType);
		}

		/// <summary>
		/// Returns a DirectoryEntry pointing to the Configuration naming context of the forest
		/// </summary>
		public DirectoryEntry ConfigurationNamingContext 
		{
			get 
			{
				if (_ConfigurationNamingContext == null) 
				{
					_ConfigurationNamingContext = GetDirectoryEntry((string) RootDse.Properties["configurationNamingContext"][0]);
				}
				return _ConfigurationNamingContext;
			}
		}

		/// <summary>
		/// Returns a DirectoryEntry pointing to the RootDSE of the queried Active Directory forest.
		/// </summary>
		public DirectoryEntry RootDse
		{
			get 
			{
				if (_RootDse == null) 
				{
					_RootDse = GetDirectoryEntry("RootDSE");
				}
				return _RootDse;
			}
		}

		/// <summary>
		/// Returns the default SMTP domain for the specified Exchange organization
		/// </summary>
		/// <param name="organization">The name of the Exchange organization</param>
		/// <returns>The default SMTP domain for the organization</returns>
		public string GetDefaultSMTPDomain(string organization) 
		{
			DirectoryEntry entry;

			entry = GetDirectoryEntry(string.Format("CN={0}, CN=Microsoft Exchange, CN=Services, {1}", organization, ConfigurationNamingContext.Properties["distinguishedName"]));

			return GetDefaultSMTPDomain(entry);
		}

		/// <summary>
		/// Returns the URI for the specified virtual directory entry.
		/// </summary>
		/// <param name="virtualDirectory">A DirectoryEntry object pointing to a virtual directory on an Exchange server</param>
		/// <returns></returns>
		public string GetUriForVirtualDirectory(DirectoryEntry virtualDirectory) 
		{
			string path;
			DirectoryEntry entry;
			string[] serverBinding;
			string connectionPath;
			PropertyValueCollection networkAddress;
			int i;
			string s;
			string serverName;
			string serverPort;

			serverBinding = null;

			if (virtualDirectory.Properties.Contains("msExchServerBindings")) 
			{
				serverBinding = ((string) virtualDirectory.Properties["msExchServerBindings"][0]).Split(new char[] {':'});
				path = "";
			}
			else 
			{
				entry = virtualDirectory;
				path = "";

				while (!entry.Properties.Contains("msExchServerBindings")) 
				{ 
					path = string.Concat(entry.Name.Substring(3), "/", path);
					entry = entry.Parent;
				}
				serverBinding = ((string) entry.Properties["msExchServerBindings"][0]).Split(new char[] {':'});
			}
			serverName = (serverBinding[2] != "") ? serverBinding[2] : serverBinding[0];
			serverPort = serverBinding[1];

			if (serverName == "") 
			{
				entry = virtualDirectory.Parent;
				while ((string) entry.Properties["name"][0] != "Protocols") entry = entry.Parent;
				entry = entry.Parent;

				networkAddress = entry.Properties["networkAddress"];
				i = 0;
				while (!((string) networkAddress[i]).StartsWith("ncacn_ip_tcp")) i++;
					
				s = (string) networkAddress[i];
				serverName = s.Substring(s.IndexOf(":")+1);

				//For Exchange 2000, the process would be finished here. But unfortunately, with Exchange 2003, thinks have changed....
				if (serverName == "") 
				{
					entry = entry.Parent;
					serverName = (string) entry.Properties["name"][0];
				}

			}

			connectionPath = string.Format("http://{0}:{1}/{2}", serverName, serverPort, path);
			return connectionPath;
		}


		/// <summary>
		/// Returns the URL for the speficied public folder store
		/// </summary>
		/// <param name="publicFolderStore">The directory entry of the public folder store</param>
		/// <param name="useAdminNamespace">Specify true to use the administrative root (http://server/exadmin/admin/smtpdomain/root)</param>
		/// <returns>An url pointing to the public folder store</returns>
		public string GetUriForPublicFolderStore(DirectoryEntry publicFolderStore, bool useAdminNamespace) 
		{
			DirectorySearcher searcher;
			SearchResult result;
			string connectionPath;
			IList proxyAddresses;
			int i;
			string domain;
			DirectoryEntry informationStore;
			
			searcher = new DirectorySearcher(publicFolderStore.Parent.Parent, string.Format("(&(objectClass=protocolCfgHTTPServer)(folderPathname={0}))", (string) publicFolderStore.Properties["name"][0]), new string[] {"msExchServerBindings"}, SearchScope.Subtree);
			result = searcher.FindOne();

			if ((result == null) || (useAdminNamespace)) 
			{
				searcher = new DirectorySearcher(publicFolderStore.Parent.Parent.Parent, "(&(objectClass=protocolCfgHTTPServer)(name=Exadmin))", new string[] {"name"}, SearchScope.Subtree);
				result = searcher.FindOne();
				
				proxyAddresses = publicFolderStore.Properties["proxyAddresses"];
				i = 0;										  
				while (!((string) proxyAddresses[i]).StartsWith("SMTP:")) i++;
				
				domain = (string) proxyAddresses[i];
				domain = domain.Substring(domain.IndexOf("@")+1);
				
				informationStore = GetDirectoryEntry((string) publicFolderStore.Properties["msExchOwningPFTree"][0]);
				
				connectionPath = string.Concat(GetUriForVirtualDirectory(result.GetDirectoryEntry()), (useAdminNamespace) ? "admin/" : "", domain, "/", informationStore.Properties["name"][0]);
			}
			else
				connectionPath = GetUriForVirtualDirectory(result.GetDirectoryEntry());

			return connectionPath;
		}

		/// <summary>
		/// Returns the default SMTP domain of the first Exchange organization.
		/// </summary>
		public string DefaultSMTPDomain
		{
			get
			{
				DirectoryEntry searchRoot;
				DirectorySearcher searcher;

				searchRoot = GetDirectoryEntry(string.Format("CN=Microsoft Exchange, CN=Services, {0}", ConfigurationNamingContext.Properties["distinguishedName"].Value));
				searcher = new DirectorySearcher(searchRoot, "(objectClass=msExchOrganizationContainer)", new string[] {"distinguishedName"}, SearchScope.OneLevel);

				return GetDefaultSMTPDomain(searcher.FindOne().GetDirectoryEntry());
			}
		}

		/// <summary>
		/// Returns the default SMTP for the specified organization.
		/// </summary>
		/// <param name="organization">A directory entry containing an Exchange organization</param>
		/// <returns>The default SMTP domain for that organization</returns>
		public string GetDefaultSMTPDomain(DirectoryEntry organization) 
		{	
			DirectoryEntry entry;
			IList property;
			int i;
			
			entry = GetDirectoryEntry("CN=Default Policy,CN=Recipient Policies, " + (string) organization.Properties["distinguishedName"][0]);

			property = entry.Properties["gatewayProxy"];

			i=0;
			while ((i<property.Count) && (((string)property[i]).IndexOf("SMTP:") == -1)) i++;

			if (i< property.Count) 
			{
				return ((string) property[i]).Substring(5);
			}
			else return "";
		}
	

		/// <summary>
		/// Returns an ADSearch object containing the Global Address List
		/// </summary>
		public ADSearch GlobalAddressList 
		{
			get 
			{
				DirectoryEntry root;
				DirectoryEntry entry;
				DirectorySearcher searcher;
				
				root = GetDirectoryEntry("CN=Microsoft Exchange, CN=Services, " + ConfigurationNamingContext);
				entry = GetDirectoryEntry(root.Properties["globalAddressList"].Value.ToString());
				root = GetDirectoryEntry(RootDse.Properties["defaultNamingContext"].Value.ToString());

				searcher = new DirectorySearcher(entry.Properties["purportedSearch"].Value.ToString());
				searcher.SearchRoot = root;
				return new ADSearch("Global Address List", searcher);
			}
		}

		/// <summary>
		/// Returns an array of <see cref="ADSearch"/> objects, containing all defined address lists.
		/// </summary>
		public ADSearch[] AllAddressLists
		{
			get 
			{
				DirectoryEntry dnRoot;
				DirectoryEntry dnAddressListRoot;
				DirectorySearcher searcher;
				ArrayList allAddressLists;
				
				dnRoot = GetDirectoryEntry("CN=Microsoft Exchange, CN=Services, " + ConfigurationNamingContext);
				
				dnAddressListRoot = GetDirectoryEntry(dnRoot.Properties["addressBookRoots"].Value.ToString());
				
				dnRoot = GetDirectoryEntry(RootDse.Properties["defaultNamingContext"].Value.ToString());
				
				allAddressLists = new ArrayList();
				foreach (DirectoryEntry addressList in dnAddressListRoot.Children) 
				{
					searcher = new DirectorySearcher(addressList.Properties["purportedSearch"].Value.ToString());
					searcher.SearchScope = SearchScope.Subtree;

					allAddressLists.Add(new ADSearch(addressList.Properties["displayName"].Value.ToString(), searcher));
				}

				return (ADSearch[]) allAddressLists.ToArray(typeof(ADSearch));
			}
		}
	}
}